-- Keyring.hs: OpenPGP (RFC4880) transferable keys parsing
-- Copyright © 2012-2014  Clint Adams
-- This software is released under the terms of the Expat license.
-- (See the LICENSE file).

module Data.Conduit.OpenPGP.Keyring (
   conduitToTKs
 , conduitToTKsDropping
 , sinkKeyringMap
) where

import Control.Applicative (many, some, (<$>), (<*>), (<|>), (<*))
import Data.Conduit
import qualified Data.Conduit.List as CL
import Data.IxSet (empty, insert)
import Data.Monoid (Monoid, (<>), mconcat)

import Codec.Encryption.OpenPGP.Types
import Data.Conduit.OpenPGP.Keyring.Instances ()
import Text.ParserCombinators.Incremental.LeftBiasedLocal (concatMany, feed, feedEof, inspect, satisfy, Parser)

data Phase = MainKey | Revs | Uids | UAts | Subs | SkippingBroken
    deriving (Eq, Ord, Show)

conduitToTKs :: MonadResource m => Conduit Pkt m TK
conduitToTKs = conduitToTKs' True

conduitToTKsDropping :: MonadResource m => Conduit Pkt m TK
conduitToTKsDropping = conduitToTKs' False

fakecmAccum :: Monad m => (accum -> (accum, [b])) -> (a -> accum -> (accum, [b])) -> accum -> Conduit a m b
fakecmAccum finalizer f =
    loop
  where
    loop accum =
        await >>= maybe (Prelude.mapM_ yield (snd (finalizer accum))) go
      where
        go a = do
            let (accum', bs) = f a accum
            Prelude.mapM_ yield bs
            loop accum'

conduitToTKs' :: MonadResource m => Bool -> Conduit Pkt m TK
conduitToTKs' intolerant = CL.filter notTrustPacket =$= CL.map (:[]) =$= fakecmAccum finalizeParsing (parseAChunk (parseTK intolerant)) ([], Just (Nothing, parseTK intolerant)) =$= CL.catMaybes
    where
        notTrustPacket (TrustPkt _) = False
        notTrustPacket _ = True

parseAChunk :: (Monoid s, Show s) => Parser s r -> s -> ([(r, s)], Maybe (Maybe (r -> r), Parser s r)) -> (([(r, s)], Maybe (Maybe (r -> r), Parser s r)), [r])
parseAChunk op a ([], Nothing) = error $ "Failure before " ++ show a
parseAChunk op a (cr, Nothing) = (inspect (feed (mconcat (map snd cr) <> a) op), map fst cr)
parseAChunk _ a (cr, Just (_, p)) = (inspect (feed a p), [])

finalizeParsing :: Monoid s => ([(r, s)], Maybe (Maybe (r -> r), Parser s r)) -> (([(r, s)], Maybe (Maybe (r -> r), Parser s r)), [r])
finalizeParsing ([], Nothing) = error $ "Unexpected finalization failure"
finalizeParsing (cr, Nothing) = (([], Nothing), map fst cr)
finalizeParsing (cr, Just (_, p)) = finalizeParsing (inspect (feedEof p))

parseTK :: Bool -> Parser [Pkt] (Maybe TK)
parseTK True = publictk True <|> secrettk True
parseTK False = publictk False <|> secrettk False <|> brokentk 6 <|> brokentk 5

(.::.) = (.) . (.) . (.) . (.) . (.)

publictk, secrettk :: Bool -> Parser [Pkt] (Maybe TK)
publictk intolerant = (Just .::. TK) <$> pkpayload <*> concatMany (signature intolerant [KeyRevocationSig,SignatureDirectlyOnAKey]) <*> (if intolerant then some else many) (signeduid intolerant) <*> many (signeduat intolerant) <*> many (signedorrevokedpubsubkey intolerant)
secrettk intolerant = (Just .::. TK) <$> skpayload <*> concatMany (signature intolerant [KeyRevocationSig,SignatureDirectlyOnAKey]) <*> (if intolerant then some else many) (signeduid intolerant) <*> many (signeduat intolerant) <*> many (raworsignedorrevokedsecsubkey intolerant)

brokentk :: Int -> Parser [Pkt] (Maybe TK)
brokentk 6 = const Nothing <$> broken 6 <* many (signature False [KeyRevocationSig,SignatureDirectlyOnAKey]) <* many (signeduid False) <* many (signeduat False) <* many (signedorrevokedpubsubkey False)
brokentk 5 = const Nothing <$> broken 5 <* many (signature False [KeyRevocationSig,SignatureDirectlyOnAKey]) <* many (signeduid False) <* many (signeduat False) <* many (raworsignedorrevokedsecsubkey False)
brokentk _ = fail "Unexpected broken packet type"

pkpayload :: Parser [Pkt] (PKPayload, Maybe SKAddendum)
pkpayload = do [PublicKeyPkt p] <- satisfy isPKP
               return (p, Nothing)
    where
        isPKP [PublicKeyPkt _] = True
        isPKP _ = False

signature :: Bool -> [SigType] -> Parser [Pkt] [SignaturePayload]
signature intolerant rts = if intolerant then signature' else (signature' <|> brokensig')
    where
        signature' = do [SignaturePkt sp] <- satisfy (isSP intolerant)
                        return $! (if intolerant then id else filter isSP') [sp]
        brokensig' = const [] <$> broken 2
        isSP True [SignaturePkt sp@(SigV3 {})] = isSP' sp
        isSP True [SignaturePkt sp@(SigV4 {})] = isSP' sp
        isSP False [SignaturePkt _] = True
        isSP _ _ = False
	isSP' (SigV3 st _ _ _ _ _ _) = st `elem` rts
	isSP' (SigV4 st _ _ _ _ _ _) = st `elem` rts
	isSP' _ = False

signeduid :: Bool -> Parser [Pkt] (String, [SignaturePayload])
signeduid intolerant = do [UserIdPkt u] <- satisfy isUID
                          sigs <- concatMany (signature intolerant [GenericCert, PersonaCert, CasualCert, PositiveCert, CertRevocationSig])
                          return (u, sigs)
    where
        isUID [UserIdPkt _] = True
        isUID _ = False

signeduat :: Bool -> Parser [Pkt] ([UserAttrSubPacket], [SignaturePayload])
signeduat intolerant = do [UserAttributePkt us] <- satisfy isUAt
                          sigs <- concatMany (signature intolerant [GenericCert, PersonaCert, CasualCert, PositiveCert, CertRevocationSig])
                          return (us, sigs)
    where
        isUAt [UserAttributePkt _] = True
        isUAt _ = False

signedorrevokedpubsubkey :: Bool -> Parser [Pkt] (Pkt, [SignaturePayload])
signedorrevokedpubsubkey intolerant = do [p] <- satisfy isPSKP
                                         sigs <- concatMany (signature intolerant [SubkeyBindingSig, SubkeyRevocationSig])
                                         return (p, sigs)
    where
        isPSKP [PublicSubkeyPkt _] = True
        isPSKP _ = False

raworsignedorrevokedsecsubkey :: Bool -> Parser [Pkt] (Pkt, [SignaturePayload])
raworsignedorrevokedsecsubkey intolerant = do [p] <- satisfy isSSKP
                                              sigs <- concatMany (signature intolerant [SubkeyBindingSig, SubkeyRevocationSig])
                                              return (p, sigs)
    where
        isSSKP [SecretSubkeyPkt _ _] = True
        isSSKP _ = False

skpayload :: Parser [Pkt] (PKPayload, Maybe SKAddendum)
skpayload = do [SecretKeyPkt p ska] <- satisfy isSKP
               return (p, Just ska)
    where
        isSKP [SecretKeyPkt _ _] = True
        isSKP _ = False

broken :: Int -> Parser [Pkt] Pkt
broken t = do [bp] <- satisfy isBroken
              return bp
    where
        isBroken [BrokenPacketPkt _ a _] = t == fromIntegral a
        isBroken _ = False

sinkKeyringMap :: MonadResource m => Sink TK m Keyring
sinkKeyringMap = CL.fold (flip insert) empty
